local addonName, addonTable = ...;

NIDisplay = {
    lastUpdateTime = 0;
}

local visionTexture;
local blindTexture;

local UPDATE_INTERVALL = 1;
local DEFAULT_MIN_ALPHA = 0.0;
local DEFAULT_MAX_ALPHA = 0.6;
local NIGHT_LENGTH = 0.6;

local combatActive = false;


BINDING_HEADER_NIGHTILLUSION_BIND_HEADER   = NightIllusion_L["BIND_HEADER"]
BINDING_NAME_NIGHTILLUSION_BIND_DECREASE   = NightIllusion_L["BIND_DECREASE"]
BINDING_NAME_NIGHTILLUSION_BIND_INCREASE   = NightIllusion_L["BIND_INCREASE"]
BINDING_NAME_NIGHTILLUSION_BIND_RESET      = NightIllusion_L["BIND_AUTO"]
BINDING_NAME_NIGHTILLUSION_BIND_ILLUMINATE = NightIllusion_L["BIND_ILLUMINATE"]
BINDING_NAME_NIGHTILLUSION_BIND_DUNGEON    = NightIllusion_L["BIND_DUNGEON"]
BINDING_NAME_NIGHTILLUSION_BIND_ACTIVE     = NightIllusion_L["BIND_ACTIVE"]

----------------------------------------------------------------------------------------------------
-- EVENTS

function NIDisplay_OnLoad( frame )
    frame:RegisterEvent( "PLAYER_REGEN_DISABLED" ); -- combat enter
    frame:RegisterEvent( "PLAYER_REGEN_ENABLED" ); -- combat leave
    frame:RegisterEvent( "PLAYER_ENTERING_WORLD" ); -- Login
    
    
    visionTexture = getglobal("NIGuiVision");
    blindTexture = getglobal("NIGuiBlind");
end

function NIDisplay_OnEvent( frame, event, ... )
    local args = {...};
    
    if ( event == "PLAYER_REGEN_DISABLED" ) then
        combatActive = NIDisplay:IsActive();
        NIDisplay:SetActive( false );
        NIDisplay:HideShadows();
    elseif ( event == "PLAYER_REGEN_ENABLED" ) then
        if combatActive then
            NIDisplay:SetActive( combatActive );
        end
    elseif ( event == "PLAYER_ENTERING_WORLD" ) then
        if NightIllusionSaved == nil then
            NIDisplay:Default();
        end
    end
end

function NIDisplay:OnUpdate( frame, elapsed )
    if not NIDisplay:IsActive() then
        return;
    end
    
	local curTime = GetTime();
	
	if ( NIDisplay.lastUpdateTime == 0 ) then
		NIDisplay.lastUpdateTime = curTime;
    elseif ( curTime - NIDisplay.lastUpdateTime > UPDATE_INTERVALL ) then
        NIDisplay.lastUpdateTime = curTime;
    else
        return;
    end
    
    if NIDisplay:IsDungeonMode() then
        NIDisplay:ShowIllusions();  
    elseif IsIndoors() then
        NIDisplay:HideShadows();
    else
        if NIDisplay:IsNight() then
            NIDisplay:ShowIllusions();
            NIDisplay:AdjustAlpha(0);
        else
            NIDisplay:HideShadows();
        end
    end
end


----------------------------------------------------------------------------------------------------
-- GET/SET Variables

function NIDisplay:Default()
    NightIllusionSaved = {
        
        startTimeH = 20;
        startTimeM = 0;
        endTimeH = 6;
        endTimeM = 0;
        
        minAlpha = 0.0;
        maxAlpha = 0.7;
        
        illuminated = true;
        dungeonMode = false;
        
        active = true;
    };
end

function NIDisplay:IsActive()
    return NightIllusionSaved.active or false;
end

function NIDisplay:SetActive( on )
    if not on then 
        NIDisplay:HideShadows();
    end
    NightIllusionSaved.active = on or false;
end

function NIDisplay:IsIlluminated() 
    return NightIllusionSaved.illuminated or false;
end

function NIDisplay:SetIllumination( on )
    if on then
        NightIllusionSaved.illuminated = true;
    else
        NightIllusionSaved.illuminated = false;
    end
end

function NIDisplay:GetMinAlpha()
    if NightIllusionSaved.minAlpha >= 0 and NightIllusionSaved.maxAlpha <= 1.0 then
        return NightIllusionSaved.minAlpha;
    else
        return 0;
    end    
end

function NIDisplay:SetMinAlpha( alpha )
    if alpha < 0 then
        alpha = 0;
    elseif alpha > 1 then
        alpha = 1.0
    end
    
    if alpha > NightIllusionSaved.maxAlpha then
        NightIllusionSaved.minAlpha = NightIllusionSaved.maxalpha;
        NightIllusionSaved.maxAlpha = alpha;
    else
        NightIllusionSaved.minAlpha = alpha;
    end
end

function NIDisplay:GetMaxAlpha()
    if NightIllusionSaved.maxAlpha >= 0 and NightIllusionSaved.maxAlpha <= 1.0 then
        return NightIllusionSaved.maxAlpha;
    else
        return DEFAULT_MAX_ALPHA;
    end    
end

function NIDisplay:SetMaxAlpha( alpha )
    if alpha < 0 then
        alpha = 0;
    elseif alpha > 1 then
        alpha = 1.0
    end
    
    if alpha < NightIllusionSaved.minAlpha then
        NightIllusionSaved.maxAlpha = NightIllusionSaved.minAlpha;
        NightIllusionSaved.minAlpha = alpha;
    else
        NightIllusionSaved.maxAlpha = alpha;
    end
end

function NIDisplay:GetStartTime()
    return NightIllusionSaved.startTimeH, NightIllusionSaved.startTimeM;
end

function NIDisplay:SetStartTime(hh, mm)
    NightIllusionSaved.startTimeH = tonumber(hh);
    NightIllusionSaved.startTimeM = tonumber(mm);
end
    
function NIDisplay:GetEndTime()
    return NightIllusionSaved.endTimeH, NightIllusionSaved.endTimeM;
end

function NIDisplay:SetEndTime(hh, mm)
    NightIllusionSaved.endTimeH = tonumber(hh);
    NightIllusionSaved.endTimeM = tonumber(mm);
end

function NIDisplay:IsDungeonMode()
    return NightIllusionSaved.dungeonMode or false;
end

function NIDisplay:SetDungeonMode( on )
    if on then
        NightIllusionSaved.dungeonMode = true;
    else
        NightIllusionSaved.dungeonMode = false;
    end
end


----------------------------------------------------------------------------------------------------
-- DISPLAY functionality

function NIDisplay:HideShadows()
    visionTexture:Hide();
    blindTexture:Hide();
end

function NIDisplay:ShowIllusions()
    if NIDisplay:IsIlluminated() then
        visionTexture:Show();
        blindTexture:Hide();
    else
        blindTexture:Show();
        visionTexture:Hide();
    end
end

function NIDisplay:ShowShadows(darknessLevel) --REM?
    visionTexture:SetAlpha(darknessLevel);
    visionTexture:Show();
end

function NIDisplay:ShowBlindness(blindnessLevel) --REM?
    blindTexture:SetAlpha(blindnessLevel);
    blindTexture:Show();
end

function NIDisplay:IsNight(now, startNight, endNight)
    if startNight== nil or endNight == nil or now == nil then
        startTh, startTm = NIDisplay:GetStartTime();
        endTh, endTm = NIDisplay:GetEndTime();
    
        local hh,mm = GetGameTime();
        startNight = NIDisplay:PrettyTime(startTh, startTm);
        endNight = NIDisplay:PrettyTime(endTh, endTm);
        now = NIDisplay:PrettyTime(hh,mm)
    end
    
    if NIDisplay:CompareTimeCircular(now, startNight, endNight) >= 0 then
        return true;
    else
        return false;
    end
end

function NIDisplay:AdjustAlpha( value )
    local alpha = visionTexture:GetAlpha();
    
    if not NIDisplay:IsDungeonMode() then
        alpha = NIDisplay:CalculateAlphaShadow();
    else
        alpha = alpha + value;
    end
    NIDisplay:SetAlpha(alpha);
end

function NIDisplay:SetAlpha(alpha)
    visionTexture:SetAlpha(alpha);
    blindTexture:SetAlpha(alpha);
end

function NIDisplay:CalculateAlphaShadow(nowT, startT, endT)
    local hh,mm = GetGameTime();
    local alpha = 0;
    if NIDisplay:IsNight() or true then
        if nowT == nil or startT == nil or endT == nil then
            startT = NIDisplay:PrettyTime( NIDisplay:GetStartTime() ); 
            endT = NIDisplay:PrettyTime( NIDisplay:GetEndTime() );
            nowT = NIDisplay:PrettyTime( hh, mm );
        end
        
        local h, m = NIDisplay:CalculateTime("24:00", startT, "-");
        h, m = NIDisplay:CalculateTime(NIDisplay:PrettyTime(h,m), endT, "+");
        local nightLengthFull = NIDisplay:TransformHoursToMinutes(h,m);

        local nightLength = nightLengthFull * NIGHT_LENGTH;
        local dawnLength = floor( (nightLengthFull * (1 - NIGHT_LENGTH)) / 2 );
        local steps = 1 / dawnLength;
        
        local dawnH, dawnM = NIDisplay:TransformMinutesToHours(dawnLength);
        local dawnTime = NIDisplay:PrettyTime(dawnH, dawnM);
        
        local duskH, duskM = NIDisplay:CalculateTime(startT, dawnTime, "+");
        local dawnH, dawnM = NIDisplay:CalculateTime(endT, dawnTime, "-");
        local duskT = NIDisplay:PrettyTime(duskH, duskM);
        local dawnT = NIDisplay:PrettyTime(dawnH, dawnM);

        local startM = NIDisplay:TransformHoursToMinutes( NIDisplay:GetStartTime() );
        local endM = NIDisplay:TransformHoursToMinutes( NIDisplay:GetEndTime() );
        local nowM = NIDisplay:TransformHoursToMinutes( hh, mm );

        if ( NIDisplay:CompareTimeCircular(nowT, startT, duskT) >= 0 ) then
        -- Dusk
            local expiredH, expiredM = NIDisplay:CalculateTime(duskT, nowT, "-");
            local expired = NIDisplay:TransformHoursToMinutes(expiredH, expiredM);
            alpha = 1 - expired * steps;
        elseif ( NIDisplay:CompareTimeCircular(nowT, duskT, dawnT) >= 0 ) then
        -- Night
            alpha = 1.0;
        elseif ( NIDisplay:CompareTimeCircular(nowT, dawnT, endT) >= 0 ) then
        -- Dawn
            local expiredH, expiredM = NIDisplay:CalculateTime(endT, nowT, "-");
            local expired = NIDisplay:TransformHoursToMinutes(expiredH, expiredM);
            alpha = expired * steps;
        else
        -- Day
            alpha = 0.0;
        end
    end
    
    if alpha > NIDisplay:GetMaxAlpha() then
        return NIDisplay:GetMaxAlpha();
    elseif alpha < NIDisplay:GetMinAlpha() then
        return NIDisplay:GetMinAlpha();
    else
        return alpha;
    end
end

function NIDisplay:IncreaseShadow( step )
    NIDisplay:SetDungeonMode( true );
    NIDisplay:AdjustAlpha( step )
end

function NIDisplay:DecreaseShadow( step )
    NIDisplay:SetDungeonMode( true );
    NIDisplay:AdjustAlpha( step )
end

function NIDisplay:ResetShadow()
    NIDisplay:SetDungeonMode( false );
    NIDisplay:AdjustAlpha( 0 );
end


----------------------------------------------------------------------------------------------------
-- HELP Functions

-- returns hh, mm in 24h format; nil on error
function NIDisplay:Calc24H( timeString )
    local hh, mm;
    
    local pTime = "^%d?%d:%d%d";
    local pMeridiem  = "[ap]%.?m%.?$";
    
    local mTime = strmatch(timeString, pTime);
    local mMeridiem = strmatch(timeString, pMeridiem);
    if mTime == nil then
        return nil;
    end
    
    hh, mm = strsplit(":", mTime);
    
    hh = tonumber(hh);
    mm = tonumber(mm);
    if mm > 60 then
        hh = hh + ((mm - mod(mm, 60)) / 60);
        mm = mm - (mm - mod(mm, 60));
    end
    
    if hh < 0 or hh > 24 then
        return nil;
    end
    
    if mm < 0 or mm > 59 then
        return nil;
    end
    
    if strfind(tostring(mMeridiem), "^p") then
        hh = hh + 12;
    end
    
    return hh, mm
end

function NIDisplay:Calc12H( timeString ) --REM
    
end

function NIDisplay:PrettyTime( hh, mm, meri )
    local sTime = "";
    
    if meri ~= nil and strmatch(meri, "[ap]%.?m%.?$") then
        if hh - 12 > 0 then
            meri = " p.m.";
            hh = hh - 12;
        else
            meri = " a.m.";
        end
    elseif meri == nil then
        meri = "";
    end
    
    if mm > 59 then
        hh = hh + (mm - mod(mm, 60))/60;
        mm = mod(mm, 60);
    end
    if hh > 23 then
        hh = mod(hh, 24);
    end
    
    if mm < 10 then
        mm = "0" .. mm;
    end
    
    sTime = hh .. ":" .. mm .. "" .. meri;
    
    return sTime;
end

-- compares t1 to t2,  -1: t1 < t2; 0: t1==t2; 1: t1 > t2; nil error
function NIDisplay:CompareTime(t1, t2)
    t1 = NIDisplay:PrettyTime(NIDisplay:Calc24H( t1 ));
    t2 = NIDisplay:PrettyTime(NIDisplay:Calc24H( t2 ));
    
    if t1 == nil or t2 == nil then
        return nil
    end
    
    local h1, m1 = strsplit(":", t1);
    local h2, m2 = strsplit(":", t2);
    
    h1 = tonumber(h1);
    h2 = tonumber(h2);
    m1 = tonumber(m1);
    m2 = tonumber(m2);
        
    if h1 > h2 or ( h1 == h2 and m1 > m2 ) then
        return 1;
    elseif h1 < h2 or ( h1 == h2 and m1 < m2 ) then
        return -1;
    else
        return 0;
    end
end

-- compares t1 with [z1, z2]:  >=0: z1 <= t1 <= z2; -1: z1 > t1 < z2
function NIDisplay:CompareTimeCircular(t1, z1, z2)
    t1 = NIDisplay:PrettyTime(NIDisplay:Calc24H( t1 ));
    z1 = NIDisplay:PrettyTime(NIDisplay:Calc24H( z1 ));
    z2 = NIDisplay:PrettyTime(NIDisplay:Calc24H( z2 ));
    
    if t1 == nil or z1 == nil or z2 == nil then
        return nil
    end
        
    local th, tm = strsplit(":", t1);
    local h1, m1 = strsplit(":", z1);
    local h2, m2 = strsplit(":", z2);
    
    th = tonumber(th);
    tm = tonumber(tm);
    h1 = tonumber(h1);
    m1 = tonumber(m1);
    h2 = tonumber(h2);
    m2 = tonumber(m2);
    
    local hasMidnight = false;
    if NIDisplay:CompareTime(z1, z2) > 0 then
        hasMidnight = true;
    end
    
    if (th == h1 and tm == m1) and (th == h2 and tm == m2) then
        -- z1 == t1 == z2
        return 0;
    elseif  th == h1 and tm == m1 then
        -- t1 == z1
        return 1;
    elseif th == h2 and tm == m2  then
        -- t1 == z2
        return 2;
    elseif ( not hasMidnight and ( th > h1 or ( th == h1 and tm > m1 ) ) and ( th < h2 or ( th == h2 and tm < m2 ) ) ) 
            then
        -- z1 < t1 < z2
        return 3; 
    elseif ( hasMidnight and ( NIDisplay:CompareTimeCircular(t1, z1, "23:59") >= 0 )) then
        return 4;
    elseif ( hasMidnight and ( NIDisplay:CompareTimeCircular(t1, "00:00", z2) >= 0 )) then
        return 5;
    else
        return -1;
    end
end

function NIDisplay:TransformHoursToMinutes( hh, mm )
    return hh * 60 + mm;
end

function NIDisplay:TransformMinutesToHours( minutes )    
    local hh, mm;
    mm = minutes;
    hh = 0;
    if mm > 59 then
        hh = (mm - mod(mm, 60))/60;
        mm = mod(mm, 60);
    end
    
    if hh > 23 then
        hh = mod(hh, 24);
    end
    
    return hh, mm;
end

-- t1 op t2 = hh, mm
function NIDisplay:CalculateTime(t1, t2, op)
    local h1, m1 = NIDisplay:Calc24H( t1 );
    local h2, m2 = NIDisplay:Calc24H( t2 );
    
    local rh, rm;
    if op == "+" then
        rh = h1 + h2;
        rm = m1 + m2;
        
        rh = rh + (rm - mod(rm, 60))/60;
        rm = rm - (rm - mod(rm,60));        
    elseif op == "-" then
        rh = h1 - h2;
        rm = m1 - m2;
        
        local lostH = 0;
        while rm < 0 do
            rm = rm + 60;
            lostH = lostH + 1;
        end
        if rh < 0 then
            rh = 24 + rh;
        end
        rh = rh - lostH;
        
        rh = rh - (rm - mod(rm, 60))/60;
        rm = rm - (rm - mod(rm,60));    
    else
        return nil;
    end
    rh = mod(rh, 24);
    
    return rh, rm
end


function NIDisplay:PrintState()
    local state = "";
    if NIDisplay:IsActive() then
        state = "|cFF00FF00[ON]|r";
    else
        state = "|cFFFF0000[OFF]|r";
    end
    
    print("NightIllusion State: " .. state);
    print("    Inside: " .. tostring(IsIndoors()));
    print("    Vision: " .. tostring(visionTexture:IsVisible()) .. ", " .. tostring(visionTexture:GetAlpha()));
    print("    Blind: " .. tostring(blindTexture:IsVisible()) .. ", " .. tostring(blindTexture:GetAlpha()));
    print(" ");
    print("    MinAlpha: " .. tostring(NIDisplay:GetMinAlpha()));
    print("    MaxAlpha: " .. tostring(NIDisplay:GetMaxAlpha()));
    print("    StartNight: " .. NIDisplay:PrettyTime(NIDisplay:GetStartTime()));
    print("    EndNight: " .. NIDisplay:PrettyTime(NIDisplay:GetEndTime()));
    print(" ");
    print("    DungeonMode: " .. tostring(NIDisplay:IsDungeonMode()));
    print("    Illumination: " .. tostring(NIDisplay:IsIlluminated()) );
    print("    IsNight: " .. tostring(NIDisplay:IsNight()) );
    print("    ");
end